#include <fstream>
#include <cstdio>
#include <vector>
#include <string>
#include <argh.hh>
#include <list>
#include "autoptr"

using namespace std;

#define USE_FDOPEN 1

#define BGR24 0x42475218  // BGR24 fourcc
#define BGR16 0x42475210  // BGR16 fourcc
#define BGR15 0x4247520F  // BGR15 fourcc

static unsigned FOURCC;

static unsigned frameus = 1000000 / 25;
static vector<string> files;

static class outfile
{
    FILE *fp;
    
    class implementation
    {
        unsigned buffering;
        long filepos, realfilepos;
        long bufbeginpos;
        vector<unsigned char> buffer;
        unsigned bufpos;
        
    public:
        FILE *fp;
        implementation() : buffering(0), filepos(0), realfilepos(0), fp(NULL)
        {
        }
        ~implementation()
        {
            if(fp) fclose(fp);
        }
        void write(const void *data, unsigned size)
        {
            if(buffering)
            {
                unsigned minbufsize = bufpos + size;
                if(buffer.size() < minbufsize)
                {
                    buffer.reserve((minbufsize + 0x3FFF) & ~0x3FFF);
                    buffer.resize(minbufsize);
                }
                memcpy(&buffer[bufpos], data, size);
                bufpos += size;
                return;
            }
            if(realfilepos != filepos)
            {
                fflush(fp);
                fseek(fp, filepos, SEEK_SET);
                realfilepos = filepos;
            }
            fwrite(data, 1, size, fp);
            filepos += size;
            realfilepos = filepos;
        }
        void setpos(long pos)
        {
            if(buffering)
            {
                if(pos < bufbeginpos)
                {
                    fprintf(stderr,
                        "Error: Attempt to seek(%ld) before bufstart(%ld) when buffering\n",
                            pos, bufbeginpos);
                    exit(-1);
                }
                bufpos = pos - bufbeginpos;
                return;
            }
            filepos = pos;
        }
        long getpos() const
        {
            if(buffering) return bufbeginpos + bufpos;
            return filepos;
        }
        void StartBuffering(unsigned estimatedsize=0)
        {
            //fprintf(stderr, "Starting buffering (pos=%ld) - buflevel was %u\n", getpos(), buffering);
            if(buffering > 20)
            {
                fprintf(stderr, "Error: Too many nested buffering levels\n");
                exit(-1);
            }
            if(buffering == 0)
            {
                bufbeginpos = getpos();
                buffer.clear();
                if(estimatedsize)
                    buffer.resize(estimatedsize);
                else
                    buffer.reserve(65536);
                bufpos = 0;
            }
            ++buffering;
        }
        void EndBuffering()
        {
            //fprintf(stderr, "Ending buffering, buflevel now %u\n", buffering-1);
            if(!buffering)
            {
                fprintf(stderr, "Error: Stopped buffering when wasn't buffering\n");
                exit(-1);
            }
            --buffering;
            if(!buffering)
            {
                write(&buffer[0], buffer.size());
                setpos(bufbeginpos + bufpos);
            }
        }
    } imp;
    
public:
    void startwrite(const string &outfilename)
    {
        imp.fp = fopen(outfilename.c_str(), "wb");
        if(!imp.fp)
        {
            perror(outfilename.c_str());
            exit(errno);
        }
    }
    void startstdout(int fd=1)
    {
#if USE_FDOPEN
        if(fd != 1)
        {
            imp.fp = fdopen(fd, "wb");
        }
        else
#endif
            imp.fp = stdout;
    }
    void ds(const char sig[4]) /* Write signature */
    {
        write(sig, 4);
    }
    void dwf(unsigned dw)      /* Write uint32 */
    {
        unsigned char Buf[4];
        Buf[0] = (unsigned char)((dw      ) & 255);
        Buf[1] = (unsigned char)((dw >> 8 ) & 255);
        Buf[2] = (unsigned char)((dw >> 16) & 255);
        Buf[3] = (unsigned char)((dw >> 24) & 255);
        write(Buf, 4);
    }
    void wf(unsigned short dw) /* Write uint16 */
    {
        unsigned char Buf[2];
        Buf[0] = (unsigned char)((dw      ) & 255);
        Buf[1] = (unsigned char)((dw >> 8 ) & 255);
        write(Buf, 2);
    }
    void putc(unsigned char c)
    {
        write(&c, 1);
    }
    void write(const void *data, unsigned size)
    {
        imp.write(data, size);
    }
    long getpos() const
    {
        return imp.getpos();
    }
    void setpos(long pos)
    {
        imp.setpos(pos);
    }
    void StartBuffering(unsigned estimatedsize=0)
    {
        imp.StartBuffering(estimatedsize);
    }
    void EndBuffering()
    {
        imp.EndBuffering();
    }
} fout;

static struct audio
{
    unsigned rate, chancount, bitness;
    double spacing;
    string file;
    
    unsigned bytecount(double seconds) const
    {
        unsigned samplecount = (unsigned)(seconds * rate);
        return samplecount * chancount * (bitness/8);
    }
    double time(unsigned bytes) const
    {
        bytes /= chancount;
        bytes /= (bitness/8);
        return bytes / (double)rate;
    }
    
    FILE *fp;
    void open()
    {
        fp = NULL;
        if(file == "-")
            fp = stdin;
#if USE_FDOPEN
        else if(file.size() > 2 && file[0] == '-' && file[1] == '#')
        {
            int fd=0;
            sscanf(file.c_str()+2, "%d", &fd);
            fp = fdopen(fd, "rb");
        }
#endif
        else if(file.size())
        {
            fp = fopen(file.c_str(), "rb");
            if(!fp)
            {
                perror(file.c_str());
                file = "";
            }
        }
        if(fp && ferror(fp))
            perror(file.c_str());
    }
} audio = {44100,2,16, 0, "", NULL};

static unsigned width, height, framecount, bytesPerPixel;

static bool xflip = false;
static bool yflip = false;

class SizeHere
{
    long pos;
public:
    SizeHere()
    {
        pos = fout.getpos();
        fout.dwf(0x7FFFFFFF);
    }
    ~SizeHere()
    {
        long now = fout.getpos();
        fout.setpos(pos);
        fout.dwf(now - pos - 4);
        fout.setpos(now);
    }
};

/* AVIupdater class!
 * It's purpose:
 * When the application starts, we are immediately beginning
 * to write the file out.
 * However we don't know surely all the data in the beginning.
 * Thus this class is for writing data that is only later known.
 */

static bool KnowFOURCC = false;
static bool KnowWidth  = false;
static bool KnowHeight = false;
static bool KnowBPP    = false;

class AVIupdater: public ptrable
{
    const char taskcode;
    unsigned bytes;
    long taskpos;
    
    void Write()
    {
        unsigned value = 0;
        
        switch(taskcode)
        {
            case 1: value = width; break;
            case 2: value = height; break;
            case 3: value = width * height * bytesPerPixel; break;
            case 4: value = framecount; break;
            case 5: value = (unsigned) (width*height * 3e6 / frameus); break;
            case 6: value = FOURCC; break;
            default:
            {
                fprintf(stderr, "AVIupdater: taskcode=%d - uhh?\n", taskcode);
                exit(-1);
            }
        }
        
        if(bytes==2) fout.wf(value);
        else if(bytes==4)fout.dwf(value);
    }
    void Finish()
    {
        long waspos = fout.getpos();
        fout.setpos(taskpos);
        
        Write();
        fout.setpos(waspos);
        if(CouldKnow()) fout.EndBuffering();
    }
    bool CouldKnow() const
    {
        return taskcode == 1
            || taskcode == 2
            || taskcode == 3
            || taskcode == 5
            || taskcode == 6;
    }
public:
    AVIupdater(char code, unsigned size) : taskcode(code), bytes(size)
    {
        if(Know())
        {
            Write();
            throw (int)(code*100 + size);
        }
        else
        {
            taskpos = fout.getpos();
            if(CouldKnow()) fout.StartBuffering();
            if(size==2) fout.wf(code+(size<<8));
            else if(size==4) fout.dwf(code+(size<<16));
        }
    }
private:
    // We don't like copying
    AVIupdater(const AVIupdater &b): ptrable(b) {}
    void operator= (const AVIupdater &) {}
public:
    ~AVIupdater()
    {
        /* Know or not, finish anyway */
        Finish();
    }
    bool Know() const
    {
        bool know = false;
        switch(taskcode)
        {
            case 1: know = KnowWidth; break;
            case 2: know = KnowHeight; break;
            case 3: know = KnowWidth && KnowHeight && KnowBPP; break;
            case 5: know = KnowWidth && KnowHeight; break;
            case 6: know = KnowFOURCC; break;
        }
        return know;
    }
};
typedef autoptr<AVIupdater> AVIupdaterptr;
static list<AVIupdaterptr> tasks;

static void CheckKnow()
{
    list<AVIupdaterptr>::iterator i, j;
    for(i=tasks.begin(); i!=tasks.end(); )
        if((*i)->Know())
        {
            j = i; ++j;
            tasks.erase(i);
            i = j;
        }
        else
            ++i;
}

static void AddTask(char taskid, unsigned size)
{
    try
    {
        AVIupdaterptr up = new AVIupdater(taskid, size);
        tasks.push_front(up);
    }
    catch(int i)
    {
        fprintf(stderr, "Bounce %d\n", i);
    }
}

static void SetFOURCC(unsigned fourcc)
{
    FOURCC = fourcc;
    KnowFOURCC = true;
    //fprintf(stderr, "FourCC is %08X\n", FOURCC);
    CheckKnow();
}
static void SetGeometry(unsigned x, unsigned y)
{
    width = x;  KnowWidth = true;
    height = y; KnowHeight = true;
    fprintf(stderr, "Geometry is %ux%u\n", width, height);
    CheckKnow();
}
static void SetBPP(unsigned bpp)
{
    bytesPerPixel = bpp;
    //fprintf(stderr, "BPP is %u\n", bpp);
    KnowBPP = true;
    CheckKnow();
}

static bool isValidTGA(FILE* fp)
{
    // Check that it's an uncompressed 24-bit tga:
    fseek(fp, 2, SEEK_SET);
    int tgaType = fgetc(fp);
    fseek(fp, 16, SEEK_SET);
    int bitsPerPixel = fgetc(fp);
    if(tgaType != 2 || bitsPerPixel != 24) return false;
    return true;
}

static void CalcSize()
{
    audio.open();

    width = 0;
    height = 0;
    
    for(unsigned i=0; i<files.size(); ++i)
    {
        if(files[i] == "-")
        {
            // Don't verify a file that
            // should be read from stdin.
            
            /*setbuf(stdin, NULL);*/
            continue;
        }
        
        FILE *fp = fopen(files[i].c_str(), "rb");
        if(!fp)
        {
            perror(files[i].c_str());
            continue;
        }
        if(isValidTGA(fp))
        {
            fseek(fp, 12, SEEK_SET);
            width = fgetc(fp);  width |= fgetc(fp) << 8;
            height = fgetc(fp); height |= fgetc(fp) << 8;
            SetGeometry(width, height);
            fclose(fp);
            return;
        }
        fclose(fp);
    }
}

static void HandleTGA(const char riffhdr[4], const string &fn)
{
    FILE *fp;
    
    if(fn == "-")
        fp = stdin;
    else
    {
        fp = fopen(fn.c_str(), "rb");
        if(!fp)
        {
            perror(fn.c_str());
            return;
        }
    }

    int datapos = fgetc(fp);
    if(fn == "-" && datapos == EOF)
    {
        // End of files in stdin
        return;
    }
    
    datapos += 18;
    
    unsigned char *HdrBuf = (unsigned char *)alloca(datapos);
    
    fread(HdrBuf+1, 1, datapos-1, fp);
    
    int tgaType = HdrBuf[2];
    unsigned new_width  = HdrBuf[12];  new_width |= HdrBuf[13] << 8;
    unsigned new_height = HdrBuf[14]; new_height |= HdrBuf[15] << 8;

    int bitsPerPixel = HdrBuf[16];
    
    if(tgaType != 2)
    {
        fprintf(stderr, "%s is not an uncompressed tga.\n", fn.c_str());
        if(fn != "-")fclose(fp);
        return;
    }
    
    switch(bitsPerPixel)
    {
        case 15: SetFOURCC(BGR15); break;
        case 16: SetFOURCC(BGR16); break;
        case 24: SetFOURCC(BGR24); break;
        default:
        {
            fprintf(stderr, "%s is not 15/16/24 -bit tga.\n", fn.c_str());
            if(fn != "-")fclose(fp);
            return;
        }
    }
    
    SetBPP((bitsPerPixel+7) / 8);

    if(!width && !height)
    {
        SetGeometry(new_width, new_height);
    }
    if(new_width != width || new_height != height)
    {
        fprintf(stderr, "%s (%ux%u) had different size than the first "
                "valid image (%ux%u)\n", fn.c_str(),
                new_width, new_height, width, height);
        if(fn == "-")
        {
            char Buf[4096];
            unsigned left = new_width * new_height;
            while(left > 0)
            {
                unsigned now = left;
                if(now > sizeof Buf)now = sizeof Buf;
                fread(Buf, 1, now, stdin);
                left -= now;
            }
        }
        else
            fclose(fp);
        return;
    }
    
    fout.StartBuffering(width*height*bytesPerPixel+8);
    {
        fout.ds(riffhdr);
        SizeHere tmp;
        
        if(yflip && fn == "-")
        {
            char *Buf = new char[width*height*bytesPerPixel];
            for(unsigned y=0; y<height; ++y)
            {
                unsigned yy = height - y - 1;
                fread(&Buf[yy*width*bytesPerPixel], bytesPerPixel, width, fp);
            }
            fout.write(Buf, bytesPerPixel*height*width);
            delete[] Buf;
        }
        else
        {
            char *Buf = new char[width*bytesPerPixel];
            for(unsigned y=0; y<height; ++y)
            {
                unsigned yy = yflip ? (height - y - 1) : y;
                
                if(fn != "-")
                    fseek(fp, datapos + yy*width*bytesPerPixel, SEEK_SET);
                fread(Buf, bytesPerPixel, width, fp);
                
                if(xflip)
                    for(unsigned x=width; x-->0; )
                        fout.write(&Buf[x*bytesPerPixel], bytesPerPixel);
                else
                    fout.write(Buf, bytesPerPixel*width);
            }
            delete[] Buf;
        }
        if(fn != "-")
            fclose(fp);
    }
    fout.EndBuffering();
}

static void WriteSTRHvids()
{
    framecount = 0;
    
    fout.ds("strh");
    SizeHere tmp;
    fout.ds("vids");
    AddTask(6, 4);   //FOURCC
    fout.dwf(0);     // flags
    fout.wf(0);      // priority
    fout.wf(0);      // language
    fout.dwf(0);     // initial frames
    fout.dwf(1000000);                  // scale
    fout.dwf((unsigned)(1e12/frameus)); // rate  (put scale*fps here)
    fout.dwf(0);     // start
    fout.dwf(0);     // scale*framecount
    fout.dwf(1048576);// suggested buffer size
    fout.dwf(10000); // quality
    AddTask(3, 4);   //width*height*3
    fout.wf(0);      // left
    fout.wf(0);      // top
    AddTask(1, 2);   //right
    AddTask(2, 2);   //bottom
}
static void WriteSTRFvids()
{
    fout.ds("strf");
    SizeHere tmp;
    fout.dwf(40);    // size - ????
    AddTask(1, 4);   //width
    AddTask(2, 4);   //height
    fout.wf(1);      // planes
    fout.wf(24);     // depth
    /* compression type */
    AddTask(6, 4);   //FOURCC
    AddTask(3, 4);   // width*height*3
    fout.dwf(0);     // ?
    fout.dwf(0);     // ?
    fout.dwf(0);     // ?
    fout.dwf(0);     // ?
}
static void WriteSTRHauds()
{
    fout.ds("strh");
    SizeHere tmp;
    fout.ds("auds");
    
    fout.dwf(0);     // FOURCC for audio
    fout.dwf(0);     // Handler
    fout.dwf(0);     // flags
    fout.wf(0);      // priority
    fout.wf(0);      // language
    fout.dwf(0);     // initial frames
    fout.dwf(1);     // scale
    fout.dwf(audio.rate); // rate
    fout.dwf(0);     // start
    fout.dwf(0);     // length (audio.rate * kesto)
    fout.dwf(1048576);// suggested buffer size
    fout.dwf(0);     // quality
    fout.dwf(audio.chancount * (audio.bitness/8)); // sample size
    fout.wf(0); fout.wf(0); fout.wf(0); fout.wf(0); // frame size
}
static void WriteSTRFauds()
{
    fout.ds("strf");
    SizeHere tmp;
    fout.wf(1);               // PCM format
    fout.wf(audio.chancount); // Channel count
    fout.dwf(audio.rate);     // Rate
    fout.dwf(audio.bytecount(1)); // Samples per second
    fout.wf(audio.chancount * (audio.bitness/8)); // Block align
    fout.wf(audio.bitness);   // Bitness
    fout.wf(0);               // cbSize
}
static void WriteAVIH()
{
    unsigned streamcount = 1;
    if(audio.file.size())
        ++streamcount;

    fout.StartBuffering();
    {
        fout.ds("avih");
        SizeHere tmp;
        fout.dwf(frameus);   // microseconds per frame
        AddTask(5, 4);       //max bytes per sec
        fout.dwf(0);         // padding (0)
        fout.dwf(0);         // flags
        AddTask(4, 4);       //framecount
        fout.dwf(0);           // initial frames
        fout.dwf(streamcount); // streamcount
        fout.dwf(1048576);     // suggested buffer size
        AddTask(1, 4);       //width
        AddTask(2, 4);       //height
        fout.dwf(0);fout.dwf(0);fout.dwf(0);fout.dwf(0); // reserved
    }
    fout.EndBuffering();
}
static void WriteSTRLvids() /* Stream list */
{
    /* STRL is a "strl" LIST */
    fout.ds("LIST");
    SizeHere tmp;
    fout.ds("strl");
    WriteSTRHvids();
    WriteSTRFvids();
}
static void WriteSTRLauds() /* Stream list */
{
    /* STRL is a "strl" LIST */
    fout.ds("LIST");
    SizeHere tmp;
    fout.ds("strl");
    WriteSTRHauds();
    WriteSTRFauds();
}
static void WriteHDRL() /* Header list */
{
    fout.StartBuffering();
    {
        /* HDRL is a "hdrl" LIST */
        fout.ds("LIST");
        SizeHere tmp;
        fout.ds("hdrl");
        WriteAVIH();
        WriteSTRLvids();
        if(audio.file.size())
        {
            WriteSTRLauds();
        }
    }
    fout.EndBuffering();
}
static void WriteMOVI() /* Write chunks */
{
    /* MOVI is a "movi" LIST */
    fout.ds("LIST");
    SizeHere tmp;
    fout.ds("movi");
    
    double videotime = audio.spacing;
    for(unsigned a=0; a<files.size(); ++a)
    {
        if(audio.fp && videotime >= audio.spacing)
        {
            unsigned bytes = audio.bytecount(videotime);
            
            char *Buf = new char[bytes];
            
            int k = fread(Buf, 1, bytes, audio.fp);
            if(ferror(audio.fp))
            {
                perror(audio.file.c_str());
                fclose(audio.fp);
                audio.fp = NULL;
            }
            else if(feof(audio.fp))
            {
                fprintf(stderr, "Error: audio ended before images did.\n");
                fclose(audio.fp);
                audio.fp = NULL;
            }
            
            fout.StartBuffering(k+8);
            {
                fout.ds("01wb"); // 01 = stream index
                SizeHere tmp;
                fout.write(Buf, k);
            }
            fout.EndBuffering();
            
            delete[] Buf;
            
            videotime -= audio.time(k);
        }
        if(1)
        {
            if(files[a] == "-")
            {
                if(feof(stdin))
                    continue;
                /*
                int c = getc(stdin);
                if(c == EOF)continue;
                ungetc(c, stdin);
                */
            }
            
            HandleTGA("00dc", files[a]); // 00 = stream index
            
            if(files[a] == "-")
            {
                --a;
            }
        }
        
        videotime += frameus / 1E6;
        ++framecount;
    }
}

static void WriteAVI()
{
    fout.ds("RIFF");
    SizeHere tmp;
    fout.ds("AVI ");
    WriteHDRL();
    WriteMOVI();
    tasks.clear();
    fprintf(stderr, "Wrote %u video frame%s\n", framecount, framecount==1?"":"s");
}

bool readInputFilenames(const string& listfile)
{
    ifstream is(listfile.c_str());
    if(!is)
    {
        fprintf(stderr, "Can't open "); perror(listfile.c_str());
        return false;
    }

    string filename;

    while(is)
    {
        getline(is, filename);
        if(is.good() && filename.size() && filename[0]!='#')
        {
            while(filename.size() && filename[0]==' ') filename.erase(0,1);
            while(filename.size() && filename[filename.size()-1]==' ')
                filename.erase(filename.size()-1,1);
            if(filename.size())
                files.push_back(filename);
        }
    }

    return true;
}

int main(int argc, const char *const *argv)
{
    string outfilename;
    
    ParamHandler Argh;
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("outfile", 'o').SetString(1).SetDesc("Specify the output file.", "<filename>");
    Argh.AddLong("fps",     'f').SetFloat(23e-5, 1e6).SetDesc("Frame rate (default: 25).", "<float>");
    Argh.AddLong("xflip",   'X').SetBool().SetDesc("Flip horizontally (not default)");
    Argh.AddLong("yflip",   'Y').SetBool().SetDesc("Flip vertically (default)");
    Argh.AddLong("list",    '@').SetString(1).SetDesc("Read input file names from the given file.", "<filename>");
    Argh.AddLong("audio",   'a').SetString(1).SetDesc("Specify raw audio source (can be a pipe).", "<filename>");
    Argh.AddLong("buffer",  'u').SetInt(0,999999).SetDesc("Specify the output buffer size in kB.", "<integer>");
    Argh.AddLong("mono",    'n').SetBool().SetDesc("Audio is mono (default: stereo).");
    Argh.AddLong("byte",    'b').SetBool().SetDesc("Audio is 8-bit (default: 16-bit).");
    Argh.AddLong("rate",    'r').SetInt(18,999999).SetDesc("Specify audio rate in hertz (default: 44100).", "<rate>");
    Argh.AddLong("audiochunksize",
                            's').SetFloat(1e-3,60).SetDesc("Specify audio spacing.", "<seconds>");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is animmaker-"VERSION"\n"
                    "Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: animmaker [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software assumes the files are uncompressed rgb24 tga files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf(
                    "\n"
                    "You can also give the tga files in stdin, example:\n"
                    " cat *.tga | animmaker -ogruu.avi -\n"
                    "\n"
                    "No warranty whatsoever.\n");
                return 0;
            case 'o':
                outfilename = Argh.GetString();
                break;
            case 'X':
                xflip = Argh.GetBool();
                break;
            case 'Y':
                yflip = Argh.GetBool();
                break;
            case 'f':
                frameus = (unsigned)(1e6 / Argh.GetFloat());
                break;
            case '@':
                if(!readInputFilenames(Argh.GetString())) return -1;
                break;
            case 'a':
                if(audio.file.size())
                {
                    fprintf(stderr, "Error: Can not specify more than one audio source!\n");
                    return -1;
                }
                audio.file = Argh.GetString();
                break;
            case 'r':
                audio.rate = Argh.GetInt();
                break;
            case 'u':
                fprintf(stderr, "Warning: -u option is not in use\n");
                //outbufsize = Argh.GetInt();
                break;
            case 'm':
                audio.chancount == Argh.GetBool() ? 1 : 2;
                break;
            case 'b':
                audio.bitness = Argh.GetBool() ? 8 : 16;
                break;
            case 's':
                audio.spacing = Argh.GetFloat();
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified. Try --help.\n");
        return -1;
    }
    
    if(audio.file == "-")
        for(unsigned f=0; f<files.size(); ++f)
            if(files[f] == "-")
            {
                fprintf(stderr, "Error: You can not read both audio and video from the same stdin. Sorry :(\n");
                return -1;
            }
    
    if(!outfilename.size() || outfilename == "-")
        fout.startstdout();
#if USE_FDOPEN
    else if(outfilename.size() > 2 && outfilename[0] == '-' && outfilename[1] == '#')
    {
        int fd=1;
        sscanf(outfilename.c_str()+2, "%d", &fd);
        fout.startstdout(fd);
    }
#endif
    else
        fout.startwrite(outfilename);
    
    if(audio.file.size())
    {
        if(!audio.spacing) audio.spacing = frameus * 2 / 1E6;
        if(audio.spacing < frameus / 1E6)
        {
            fprintf(stderr, "Error: Audio spacing (%g) can not be less than video frame length (%g).\n", audio.spacing, frameus/1E6);
            return -1;
        }
        if(audio.spacing > 1)
        {
            fprintf(stderr, "Warning: Audio spacing (%g) should usually be less than second.\n", audio.spacing);
        }
    }
    
    CalcSize();
    WriteAVI();
    
    return 0;
}
